from .build import build
from .container import container
from .fetch import fetch
from .image import image
from .image_or_build import image_or_build
from .network import network
from .volume import volume


__all__ = [
    'build',
    'container',
    'fetch',
    'image',
    'image_or_build',
    'network',
    'volume',
]



from pytest_docker_tools import volume

test_volume_1 = volume()


def test_volume_created(docker_client, test_volume_1):
    for v in docker_client.volumes.list():
        if v.id == test_volume_1.id:
            # Looks like we managed to start one!
            break
    else:
        assert False, 'Looks like we failed to create a volume'



import uuid

from pytest_docker_tools.builder import fixture_factory


@fixture_factory()
def network(request, docker_client, wrapper_class, **kwargs):
    ''' Docker network '''

    name = kwargs.pop('name', 'pytest-{uuid}').format(uuid=str(uuid.uuid4()))

    print(f'Creating network {name}')
    network = docker_client.networks.create(name, **kwargs)
    request.addfinalizer(lambda: network.remove())

    wrapper_class = wrapper_class or (lambda network: network)
    return wrapper_class(network)



import sys

from pytest_docker_tools.builder import fixture_factory


@fixture_factory(scope='session')
def build(request, docker_client, wrapper_class, **kwargs):
    ''' Docker image: built from "{path}" '''

    # The docker build command now defaults to --rm=true, but docker-py doesnt
    # Let's do what docker build does by default
    kwargs.setdefault('rm', True)

    sys.stdout.write(f'Building {kwargs["path"]}')

    try:
        image, logs = docker_client.images.build(**kwargs)

        for line in logs:
            sys.stdout.write('.')
            sys.stdout.flush()

    finally:
        sys.stdout.write('\n')

    # request.addfinalizer(lambda: docker_client.images.remove(image.id))

    wrapper_class = wrapper_class or (lambda image: image)
    return wrapper_class(image)



#! /usr/bin/python

import fnmatch
import os

from dnslib import MX, PTR, RR, SOA, TXT, A
from dnslib.dns import QTYPE
from dnslib.server import BaseResolver, DNSLogger, DNSServer


class Logger(DNSLogger):

    def log_request(self, handler, request):
        print(request)

    def log_reply(self, handler, reply):
        print(reply)
        self.log_data(reply)


class Resolver(BaseResolver):

    def _resolve_request(self, request):
        name = str(request.q.qname).strip('.').replace('.', '_').upper()
        rec_type = QTYPE[request.q.qtype]
        environ_key = f'DNS_{name}__{rec_type}'

        for glob, value in os.environ.items():
            if not glob.startswith('DNS_'):
                continue

            if fnmatch.fnmatch(environ_key, glob):
                return value

    def resolve(self, request, handler):
        reply = request.reply()
        value = self._resolve_request(request)

        if value and request.q.qtype == QTYPE.MX:
            reply.add_answer(
                RR(request.q.qname, QTYPE.MX, rdata=MX(label=value)),
            )
        elif value and request.q.qtype == QTYPE.A:
            reply.add_answer(
                RR(request.q.qname, QTYPE.A, rdata=A(value)),
            )
        elif value and request.q.qtype == QTYPE.TXT:
            reply.add_answer(RR(
                request.q.qname,
                QTYPE.TXT,
                rdata=TXT(value),
            ))
        elif value and request.q.qtype == QTYPE.PTR:
            reply.add_answer(RR(
                request.q.qname,
                QTYPE.PTR,
                rdata=PTR(value),
            ))

        reply.add_auth(
            RR(request.q.qname, QTYPE.SOA, ttl=60, rdata=SOA(
                "ns1.local.dev", "ns2.local.dev",
                (20140101, 3600, 3600, 3600, 3600),
            ))
        )

        reply.add_ar(RR("ns1.local.dev", ttl=60, rdata=A("127.0.0.1")))

        return reply


if __name__ == '__main__':
    logger = Logger()

    server = DNSServer(
        Resolver(),
        port=53,
        address='0.0.0.0',
        logger=logger,
    )

    server.start()



import os
import socket

from pytest_docker_tools import build, container, fetch
from pytest_docker_tools.utils import wait_for_callable

test_container_1_image = fetch(repository='redis:latest')

test_container_1 = container(
    image='{test_container_1_image.id}',
    ports={
        '6379/tcp': None,
    },
)

ipv6_folder = os.path.join(os.path.dirname(__file__), 'fixtures/ipv6')
ipv6_image = build(path=ipv6_folder)
ipv6 = container(
    image='{ipv6_image.id}',
    ports={
        '1234/udp': None,
    }
)


def test_container_created(docker_client, test_container_1):
    for c in docker_client.containers.list(ignore_removed=True):
        if c.id == test_container_1.id:
            # Looks like we managed to start one!
            break
    else:
        assert False, 'Looks like we failed to start a container'


def test_container_ipv6(ipv6):
    addr = ipv6.get_addr('1234/udp')
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.sendto(b'msg', addr)

    wait_for_callable('Waiting for delivery confirmation', lambda: 'msg' in ipv6.logs())



'''
# Template handling

Fixture factories can take static strings:

```python
redis = container(
    image='redis:latest',
)
```

But that is not useful when building multiple containers that need to reference
one another or you need to parameterize the fixture. This module provides two
facilities: The ability to reference fixtures using python string template
notation and the ability to know what fixtures this will fetch in at test
collection (and generation) time.

For example:

```
def test_simple_resolve(request):
    # These are parameters declared at import time - they can be reevaluated in the context of multiple tests
    kwargs = {
        'somekey': ['{pytestconfig.getoption("verbose")}'],
    }

    # This can be used in a fixture factory to fill in the templates
    resolved = resolve_fixtures_in_params(request, kwargs)

    # And then the test can access them
    pytestconfig = request.getfixturevalue('pytestconfig')
    assert resolved['somekey'][0] == str(pytestconfig.getoption("verbose"))
}
```

In order to make fixtures generated by a fixture factory more seamless we need
to know a fixtures dependencies at collection time. We have a helper to find
them:

def test_simple_find():
    # These are parameters declared at import time - they can be reevaluated in the context of multiple tests
    kwargs = {
        'somekey': ['{pytestconfig.getoption("verbose")}'],
    }

    dependencies = find_fixtures_in_params(kwargs)
    assert dependencies = set('pytestconfig')
'''

import inspect
from string import Formatter

__all__ = [
    'find_fixtures_in_params',
    'resolve_fixtures_in_params',
]


class FixtureFormatter(Formatter):

    def __init__(self, request):
        self.request = request

    def get_value(self, key, args, kwargs):
        return self.request.getfixturevalue(key)


class Renderer(object):

    def __init__(self, request):
        self.request = request

    def visit_value(self, val):
        if isinstance(val, str):
            return FixtureFormatter(self.request).format(val)
        elif callable(val):
            return val(*[self.request.getfixturevalue(f) for f in inspect.getargspec(val)[0]])
        return val

    def visit_list(self, val):
        return [self.visit(v) for v in val]

    def visit_dict(self, mapping):
        return {self.visit(k): self.visit(v) for (k, v) in mapping.items()}

    def visit(self, value):
        if isinstance(value, dict):
            return self.visit_dict(value)
        elif isinstance(value, list):
            return self.visit_list(value)
        elif value:
            return self.visit_value(value)


class FixtureFinder(object):

    def visit_value(self, val):
        if isinstance(val, str):
            for literal_text, format_spec, conversion, _ in Formatter().parse(val):
                if format_spec:
                    yield format_spec.split('.')[0].split('[')[0]
        elif callable(val):
            yield from inspect.getargspec(val)[0]

    def visit_list(self, val):
        for v in val:
            yield from self.visit(v)

    def visit_dict(self, mapping):
        for k, v in mapping.items():
            yield from self.visit(k)
            yield from self.visit(v)

    def visit(self, value):
        if isinstance(value, dict):
            yield from self.visit_dict(value)
        elif isinstance(value, list):
            yield from self.visit_list(value)
        elif value:
            yield from self.visit_value(value)


def find_fixtures_in_params(value):
    '''
    Walk an object and identify fixtures references in templates in strings.
    '''
    finder = FixtureFinder()
    return set(finder.visit(value))


def resolve_fixtures_in_params(request, value):
    '''
    Walk an object and resolve fixture values referenced in template strings.
    '''
    renderer = Renderer(request)
    return renderer.visit(value)



import docker
import pytest

from .exceptions import ContainerError
from .wrappers import Container


@pytest.fixture(scope='session')
def docker_client(request):
    ''' A Docker client configured from environment variables '''
    return docker.from_env()


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    '''
    This hook allows Docker containers to contribute their logs to the py.test
    report.
    '''

    outcome = yield
    rep = outcome.get_result()

    if not rep.failed:
        return

    if call.excinfo and isinstance(call.excinfo.value, ContainerError):
        container = call.excinfo.value._container
        rep.sections.append((container.name, container.logs()))

    if 'request' not in item.funcargs:
        return

    for name, fixturedef in item.funcargs['request']._fixture_defs.items():
        if not hasattr(fixturedef, 'cached_result'):
            continue
        fixture = fixturedef.cached_result[0]
        if isinstance(fixture, Container):
            rep.sections.append((
                name + ': ' + fixture.name,
                fixture.logs(),
            ))



import os
import sys
import time

from .exceptions import TimeoutError


def wait_for_callable(message, callable, timeout=30):
    '''
    Runs a callable once a second until it returns True or we hit the timeout.
    '''
    sys.stdout.write(message)
    try:
        for i in range(timeout):
            sys.stdout.write('.')
            sys.stdout.flush()

            if callable():
                return

            time.sleep(1)
    finally:
        sys.stdout.write('\n')

    raise TimeoutError(f'Timeout of {timeout}s exceeded')


def tests_inside_container():
    ''' Returns True if tests are running inside a Linux container '''

    if not os.path.exists('/proc/1/sched'):
        return False

    with open('/proc/1/sched', 'r') as fp:
        line1 = fp.read().split('\n')[0]

    # Right now this file contains a header like this which leaks the actual pid
    #     systemd (1, #threads: 1)
    # If its not '1' we have detected containment

    init, info = line1.split(' ', 1)
    pid, threads = info.strip('(').rstrip(')').split(', ', 1)
    return pid != '1'



from pytest_docker_tools import network

test_network_1 = network()


def test_network_created(docker_client, test_network_1):
    for n in docker_client.networks.list():
        if n.id == test_network_1.id:
            # Looks like we managed to start one!
            break
    else:
        assert False, 'Looks like we failed to create a network'



class TimeoutError(Exception):
    pass


class ContainerError(Exception):

    def __init__(self, container, *args, **kwargs):
        self._container = container
        super().__init__(*args, **kwargs)


class ContainerFailed(ContainerError):
    pass


class ContainerNotReady(ContainerError):
    pass



#! /usr/bin/python

import json
import socket
from http.server import BaseHTTPRequestHandler, HTTPServer


class RequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        result = {
            'result': socket.gethostbyname('example.com'),
        }

        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(result).encode('utf-8'))


if __name__ == '__main__':
    try:
        server = HTTPServer(('', 8080), RequestHandler)
        server.serve_forever()
    except KeyboardInterrupt:
        pass



import io
import os
import tarfile
import uuid

from pytest_docker_tools.builder import fixture_factory


def _populate_volume(docker_client, volume, seeds):
    fp = io.BytesIO()
    tf = tarfile.open(mode="w:gz", fileobj=fp)

    for path, contents in seeds.items():
        ti = tarfile.TarInfo(path)
        if contents is None:
            ti.type = tarfile.DIRTYPE
            tf.addfile(ti)
        else:
            ti.size = len(contents)
            tf.addfile(ti, io.BytesIO(contents))

    tf.close()
    fp.seek(0)

    image, logs = docker_client.images.build(
        path=os.path.join(os.path.dirname(__file__), '..', 'contexts/scratch'),
        rm=True,
    )
    list(logs)

    container = docker_client.containers.create(
        image=image.id,
        volumes={
            f'{volume.name}': {'bind': '/data'},
        },
    )

    try:
        container.put_archive('/data', fp)
    finally:
        container.remove(force=True)


@fixture_factory()
def volume(request, docker_client, wrapper_class, **kwargs):
    ''' Docker volume '''

    name = kwargs.pop('name', 'pytest-{uuid}').format(uuid=str(uuid.uuid4()))
    seeds = kwargs.pop('initial_content', {})

    print(f'Creating volume {name}')
    volume = docker_client.volumes.create(name, **kwargs)
    request.addfinalizer(lambda: volume.remove(True))

    if seeds:
        _populate_volume(docker_client, volume, seeds)

    wrapper_class = wrapper_class or (lambda volume: volume)
    return wrapper_class(volume)



import json


def test_api_server(apiclient):
    apiclient.request('GET', '/')
    response = apiclient.getresponse()
    assert response.status == 200
    assert json.loads(response.read()) == {'result': '127.0.0.1'}



from pytest_docker_tools.builder import fixture_factory
from pytest_docker_tools.exceptions import ContainerNotReady, TimeoutError
from pytest_docker_tools.utils import wait_for_callable
from pytest_docker_tools.wrappers import Container


@fixture_factory()
def container(request, docker_client, wrapper_class, **kwargs):
    ''' Docker container: image={image} '''

    timeout = kwargs.pop('timeout', 30)

    kwargs.update({'detach': True})

    raw_container = docker_client.containers.run(**kwargs)
    request.addfinalizer(lambda: raw_container.remove(force=True) and raw_container.wait(timeout=10))

    wrapper_class = wrapper_class or Container
    container = wrapper_class(raw_container)

    try:
        wait_for_callable('Waiting for container to be ready', container.ready, timeout)
    except TimeoutError:
        raise ContainerNotReady(container, 'Timeout while waiting for container to be ready')

    return container



'''
An opionated set of helpers for defining Docker integration test environments with py.test fixtures.
'''

from .factories import build, container, fetch, image, image_or_build, network, volume


__version__ = '0.2.2'


__all__ = [
    'build',
    'container',
    'fetch',
    'image',
    'image_or_build',
    'network',
    'volume',
]



import textwrap

from .templates import find_fixtures_in_params, resolve_fixtures_in_params


def build_fixture_function(callable, scope, wrapper_class, kwargs):
    name = callable.__name__
    docstring = getattr(callable, '__doc__', '').format(**kwargs)
    fixtures = find_fixtures_in_params(kwargs).union(set(('request', 'docker_client')))
    fixtures_str = ','.join(fixtures)

    template = textwrap.dedent(f'''
    import pytest

    @pytest.fixture(scope=scope)
    def {name}({fixtures_str}):
        \'\'\'
        {docstring}
        \'\'\'
        real_kwargs = resolve_fixtures_in_params(request, kwargs)
        return _{name}(request, docker_client, wrapper_class=wrapper_class, **real_kwargs)
    ''')
    globals = {
        'resolve_fixtures_in_params': resolve_fixtures_in_params,
        f'_{name}': callable,
        'kwargs': kwargs,
        'scope': scope,
        'wrapper_class': wrapper_class,
    }
    exec(template, globals)
    return globals[name]


def fixture_factory(scope='function'):
    wrapper_class = None

    def inner(callable):
        def factory(*, scope=scope, wrapper_class=wrapper_class, **kwargs):
            return build_fixture_function(callable, scope, wrapper_class, kwargs)
        factory.__name__ = callable.__name__
        factory.__doc__ = getattr(callable, '__doc__', '')
        return factory
    return inner



'''
This module contains tests of the 'orchestration' of dependent docker units.
If a container depends on an image and a container, and that container depends
on another container, and so on, then all the contains should be built in the
right order.
'''

from pytest_docker_tools import build, container, fetch, network, volume

redis_image = fetch(repository='redis:latest')

redis0 = container(
    image='{redis_image.id}',
    environment={
        'MARKER': 'redis0-0sider',
    }
)

foobar = build(path='tests/integration')
mynetwork = network()
myvolume = volume()
mycontainer = container(
    image='{foobar.id}',
    network='{mynetwork.id}',
    volumes={
        '{myvolume.id}': {'bind': '/var/tmp'},
    },
    environment={
        'REDIS_IP': lambda redis0: redis0.ips.primary,
    },
    dns=['{redis0.ips.primary}'],
)


def test_related_container_created(docker_client, mycontainer):
    ''' Creating mycontainer should pull in redis0 because we depend on it to calculate an env variable '''
    backend_ip = mycontainer.env['REDIS_IP']
    for c in docker_client.containers.list(ignore_removed=True):
        if 'MARKER=redis0-0sider' in c.attrs['Config']['Env'] and backend_ip in str(c.attrs):
            break
    else:
        assert False, 'redis0 not running'


def test_gets_related_container_ip(redis0, mycontainer):
    ''' The lambda we passed to environment should have been executed with the redis fixture value '''
    assert mycontainer.env['REDIS_IP'] == redis0.ips.primary


def test_gets_volume(myvolume, mycontainer):
    ''' The container should have a volume configured pointing at our fixturized volume '''
    for mount in mycontainer.attrs['Mounts']:
        if mount['Name'] == myvolume.name:
            break
    else:
        assert False, 'Could not find attached volume'


def test_gets_network(mynetwork, mycontainer):
    ''' The container should be attached to our fixturized network '''
    assert mynetwork.name in mycontainer.ips


def test_gets_dns_ip(redis0, mycontainer):
    ''' The container should be attached to our fixturized network '''
    assert redis0.ips.primary == mycontainer.attrs['HostConfig']['Dns'][0]



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pytest_docker_tools',
 'pytest_docker_tools.factories',
 'pytest_docker_tools.wrappers']

package_data = \
{'': ['*'], 'pytest_docker_tools': ['contexts/*', 'contexts/scratch/*']}

install_requires = \
['pytest', 'docker']

entry_points = \
{'pytest11': ['docker_tools = pytest_docker_tools.plugin']}

setup(name='pytest_docker_tools',
      version='0.2.2',
      description='An opionated set of helpers for defining Docker integration test environments with py.test fixtures.',
      author='John Carr',
      author_email='john.carr@unrouted.co.uk',
      url='https://github.com/Jc2k/pytest-docker-tools',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



import sys

from pytest_docker_tools.builder import fixture_factory


@fixture_factory(scope='session')
def fetch(request, docker_client, wrapper_class, **kwargs):
    ''' Docker image: Fetched from {repository} '''

    sys.stdout.write(f'Fetching {kwargs["repository"]}\n')

    image = docker_client.images.pull(**kwargs)
    # request.addfinalizer(lambda: docker_client.images.remove(image.id))

    wrapper_class = wrapper_class or (lambda image: image)
    return wrapper_class(image)



from pytest_docker_tools.builder import fixture_factory


@fixture_factory(scope='session')
def image(request, docker_client, wrapper_class, **kwargs):
    ''' Docker image: named "{name}" (already available) '''

    image = docker_client.images.get(kwargs['name'])
    wrapper_class = wrapper_class or (lambda image: image)
    return wrapper_class(image)



import os
from http.client import HTTPConnection

import pytest
from pytest_docker_tools import build, container

fakedns_image = build(
    path=os.path.join(os.path.dirname(__file__), 'dns'),
)

fakedns = container(
    image='{fakedns_image.id}',
    environment={
        'DNS_EXAMPLE_COM__A': '127.0.0.1',
    }
)

apiserver_image = build(
    path=os.path.join(os.path.dirname(__file__), 'api'),
)


apiserver = container(
    image='{apiserver_image.id}',
    ports={
        '8080/tcp': None,
    },
    dns=['{fakedns.ips.primary}']
)


@pytest.fixture
def apiclient(apiserver):
    port = apiserver.ports['8080/tcp'][0]
    return HTTPConnection(f'localhost:{port}')
